-- Support functions for CalcStat. These consist of implementations of more complex calculation types and rounding/min/max/compare functions for floating point numbers.
-- Created by Giseldah

-- Floating point numbers bring errors into the calculation, both inside the Lotro-client and in this function collection. This is why a 100% match with the stats in Lotro is impossible.
-- Anyway, to compensate for some errors, we use a calculation deviation correction value. This makes for instance 24.49999999 round to 25, as it's assumed that 24.5 was intended as outcome of a formula.
DblCalcDev = 0.00000001;

-- Armour Multipliers for CalcArm function
local ArmItemTypeMP = {9,9,18.75,30,15,23.75,12,37.5};
local ArmQualityCommonMP = {220/3,220/3,80,80,76,80.4,80,70};
local ArmQualityUncommonMP = {220/3,220/3,85,85,82,85.2,85,75};
local ArmQualityRareMP = {260/3,260/3,90,90,88,90,90,80};
local ArmQualityIncomparableMP = {280/3,280/3,95,95,94,94.8,95,95};
local ArmQualityLegendaryMP = {100,100,100,100,100,100.2,100,100};
local ArmQualityMP = {};
local ArmCategoryMP = {22.8877/19,19.1/19,15.2/19};
local ArmPointAMP = {1,1/75,0.0132985,0.0132863};
local ArmPointBMP = {79,1,0.9973834,0.9964283};
local ArmPointCMP = {80,1.1034482,1.0994843,1.1};
local ArmPointDMP = {200,2,2,2};
local ArmPointEMP = {225,3,3,3};
local ArmPointFMP = {300,3.6,3.6,3.6};
local ArmPointGMP = {349,6,6,6};
local ArmPointHMP = {350,6.6,6.6,6.6};
local ArmPointIMP = {399,15,15,15};
local ArmPointMP = {};

-- Roman number conversion data for RomanRankDecode function
local RomanCharsToValues = {["M"]=1000,["CM"]=900,["D"]=500,["CD"]=400,["C"]=100,["XC"]=90,["L"]=50,["XL"]=40,["X"]=10,["IX"]=9,["V"]=5,["IV"]=4,["I"]=1};

-- Calc(ulate)Arm(our): calculates armour value based on a type code and the item level.
-- Armour calculation is based on multiplying values for type, quality, category and item level graph points which leads to a composed graph of straight lines: determine the right line and calculate the item level point on it.
-- acode string:
-- 1st position: H=heavy, M=medium, L=light
-- 2nd position: H=head, S=shoulders, CL=cloak/back, C=chest, G=gloves, L=leggings, B=boots, Sh=shield
-- 3rd position: W=white/common, Y=yellow/uncommon, P=purple/rare, T=teal/blue/incomparable, G=gold/legendary/epic
-- Note: no such thing exists as a heavy, medium or light cloak, so no H/M/L in cloak codes (cloaks go automatically in the M class since U23, although historically this was L)
-- This function has been tested with all armour gear in the game and should give a 100% match unless the item is very old and uses old system values (maybe calculated or just constants).
-- Also, with many updates/expansions some mistakes have been made with asigning the right quality to gear. It might be that a gold shield only has a blue armour value for instance. I can't recall that any mistake (or intentional assignment) has been made more than one quality category in difference.

function CalcArm(acode, ilvl)

	local armourcode = string.upper(string.match(acode,"(%a+)"));

	-- get positional codes and make some corrections
	local sarmclass = string.sub(armourcode,1,1);
	local sarmtype = string.sub(armourcode,2,2);
	local sarmcol = string.sub(armourcode,3,3);
	if sarmtype == "S" and sarmcol == "H" then
		sarmtype = "SH";
		sarmcol = string.sub(armourcode,4,4);
	elseif sarmclass == "C" and sarmtype == "L" then
		sarmclass = "M";
		sarmtype = "CL";
	else
		sarmtype = " "..sarmtype;
	end
	
	-- convert codes to indexes for MP tables
	local iarmclass = string.find("HML",sarmclass);
	local iarmtype = (string.find(" H SCL C G L BSH",sarmtype)+1)/2;
	local iarmcol = string.find("WYPTG",sarmcol);
	
	if ilvl > 0 and ilvl <= 9999 and iarmclass >= 1 and iarmtype >= 1 and iarmcol >= 1 then
		-- construct some MP tables if not done already
		if #ArmQualityMP == 0 then
			table.insert(ArmQualityMP,ArmQualityCommonMP);
			table.insert(ArmQualityMP,ArmQualityUncommonMP);
			table.insert(ArmQualityMP,ArmQualityRareMP);
			table.insert(ArmQualityMP,ArmQualityIncomparableMP);
			table.insert(ArmQualityMP,ArmQualityLegendaryMP);
		end
		if #ArmPointMP == 0 then
			table.insert(ArmPointMP,ArmPointAMP);
			table.insert(ArmPointMP,ArmPointBMP);
			table.insert(ArmPointMP,ArmPointCMP);
			table.insert(ArmPointMP,ArmPointDMP);
			table.insert(ArmPointMP,ArmPointEMP);
			table.insert(ArmPointMP,ArmPointFMP);
			table.insert(ArmPointMP,ArmPointGMP);
			table.insert(ArmPointMP,ArmPointHMP);
			table.insert(ArmPointMP,ArmPointIMP);
		end

		-- determine the start and end point of the graph segment: ilvl_start <= ilvl <= ilvl_end
		-- pi_start, pi_end will never be 1/1 or last/last, but atmost 1/2 or last-1/last, so that ilvl outside ilvl_first/ilvl_last can still be calculated by extending the outer line (like ilvl 400 currently with ilvl_last=399).
		local pi_start = 1;
		local pi_end = #ArmPointMP;
		for pi, pmp in ipairs(ArmPointMP) do
			if pmp[1] >= ilvl and pi > 1 then
				pi_end = pi;
				break;
			end
			if pmp[1] <= ilvl and pi < #ArmPointMP then
				pi_start = pi;
			end
		end

		-- round down the resulting value on the graph line
		return FloorDbl( LinFmod(	ArmItemTypeMP[iarmtype] * ArmQualityMP[iarmcol][iarmtype] * ArmCategoryMP[iarmclass],
					ArmPointMP[pi_start][1+iarmclass],
					ArmPointMP[pi_end][1+iarmclass],
					ArmPointMP[pi_start][1],
					ArmPointMP[pi_end][1],
					ilvl ) );
	else
		return 0;
	end

end

-- LinFmod: Linear line function between 2 points with some optional modifications.
-- Connects point (dLstart,dVal*dFstart) with (dLend,dVal*dFend).
-- Usually used with dVal=1 and dFstart/dFend containing unrelated points or dVal=# and dFstart/dFend containing multiplier factors.
-- Modification for in-between points on the line: rounding (in general to multiples of dRound=10).
-- Related to Calculation Type "T" (Trap=staircase in Dutch). _-

function LinFmod(dVal, dFstart, dFend, dLstart, dLend, dLvl, dRound)

	if IsDbl(dLvl,dLstart) then
		return dVal*dFstart;
	elseif IsDbl(dLvl,dLend) then
		return dVal*dFend;
	elseif dRound == nil then
		return dVal*(dFstart*(dLend-dLvl)+(dLvl-dLstart)*dFend)/(dLend-dLstart);
	else
		return RoundDbl((dVal*(dFstart*(dLend-dLvl)+(dLvl-dLstart)*dFend)/((dLend-dLstart)*dRound)))*dRound;
	end

end

-- ExpFmod: Exponential function based on percentage.
-- Common percentage values are around ~5.5% for between levels and ~20% jumps between level segments.
-- Related to Calculation Type "E".

function ExpFmod(dVal, dLstart, dPlvl, dLvl)

	if SmallerDbl(dLvl,dLstart) then
		return dVal;
	else
		return dVal*(1+dPlvl/100)^(dLvl-dLstart+1);
	end

end

-- CalcPercAB: Calculates the percentage out of a rating based on the AB formula.
-- Related to Calculation Type "P".

function CalcPercAB(dA, dB, dPCap, dR)

	if dR <= 0 then
		return 0;
	else
		return MinDbl(dA/(1+dB/dR),dPCap);
	end

end

-- CalcRatAB: Calculates the rating out of a percentage based on the AB formula.
-- Related to Calculation Type "R".

function CalcRatAB(dA, dB, dCapR, dP)

	if dP <= 0 then
		return 0;
	else
		return MinDbl(dB/(dA/dP-1),dCapR);
	end
end

-- NamedRangeValue: Takes a value from a named spreadsheet table.
-- This function doesn't have a meaning outside a spreadsheet and so is not implemented here.
-- Related to Calculation Type "N".

function NamedRangeValue(RName, RowIndex, ColIndex)

	return 0;

end

-- Misc. functions for floating point: rounding etc.
-- For roundings: iDec is number of decimals.

function RoundDbl(dNum, iDec)

	if iDec == nil then
		iDec = 0;
	end
	if iDec == 0 then
		return math.floor(dNum+0.5+DblCalcDev);
	else
		return math.floor(dNum*10^iDec+0.5+DblCalcDev)/10^iDec;
	end
	
end

function CeilDbl(dNum, iDec)

	if iDec == nil then
		iDec = 0;
	end
	if iDec == 0 then
		return math.floor(dNum+1-DblCalcDev);
	else
		return math.floor(dNum*10^iDec+1-DblCalcDev)/10^iDec;
	end
	
end

function FloorDbl(dNum, iDec)

	if iDec == nil then
		iDec = 0;
	end
	if iDec == 0 then
		return math.floor(dNum+DblCalcDev);
	else
		return math.floor(dNum*10^iDec+DblCalcDev)/10^iDec;
	end

end

function IsDbl(dNum1, dNum2)

	return (math.abs(dNum1-dNum2) <= DblCalcDev);

end

function SmallerDbl(dNum1, dNum2)

	return ((dNum2-dNum1) > DblCalcDev);

end

function GreaterDbl(dNum1, dNum2)

	return ((dNum1-dNum2) > DblCalcDev);

end

function MinDbl(dNum1, dNum2)

	if dNum1 <= dNum2 then
		return dNum1;
	else
		return dNum2;
	end

end

function MaxDbl(dNum1, dNum2)

	if dNum1 >= dNum2 then
		return dNum1;
	else
		return dNum2;
	end

end

-- RomanRankDecode: converts a string with a Roman number in characters, to an integer number
-- used for Legendary Item Title calculation.

function RomanRankDecode(srank)

	if srank ~= "" then
		for schars, ivalue in pairs(RomanCharsToValues) do
			if string.sub(string.upper(srank),1,string.len(schars)) == schars then
				return ivalue+RomanRankDecode(string.sub(srank,string.len(schars)+1));
			end
		end
	end

	return 0;
	
end
